#!/usr/bin/env pypy -u
# -*- coding: utf-8 -*-

from __future__ import print_function
from __future__ import division
import importlib
import re
import sys
import time
import logging
import argparse

import tools
import sunfish

from tools import WHITE, BLACK, Unbuffered

def main():
    parser = argparse.ArgumentParser()
    parser.add_argument('module', help='sunfish.py file (without .py)', type=str, default='sunfish', nargs='?')
    parser.add_argument('--tables', metavar='pst', help='alternative pst table', type=str, default=None)
    args = parser.parse_args()

    sunfish = importlib.import_module(args.module)
    if args.tables is not None:
        pst_module = importlib.import_module(args.tables)
        sunfish.pst = pst_module.pst

    logging.basicConfig(filename='sunfish.log', level=logging.DEBUG)
    out = Unbuffered(sys.stdout)
    def output(line):
        print(line, file=out)
        logging.debug(line)
    pos = tools.parseFEN(tools.FEN_INITIAL)
    searcher = sunfish.Searcher()
    color = WHITE
    our_time, opp_time = 1000, 1000 # time in centi-seconds
    show_thinking = True

    stack = []
    while True:
        if stack:
            smove = stack.pop()
        else:
            smove = input()

        logging.debug(f'>>> {smove} ')

        if smove == 'quit':
            break

        elif smove == 'uci':
            output('id name Sunfish')
            output('id author Thomas Ahle & Contributors')
            output('uciok')

        elif smove == 'isready':
            output('readyok')

        elif smove == 'ucinewgame':
            stack.append('position fen ' + tools.FEN_INITIAL)

        # syntax specified in UCI
        # position [fen  | startpos ]  moves  ....

        elif smove.startswith('position'):
            params = smove.split(' ')
            idx = smove.find('moves')

            if idx >= 0:
                moveslist = smove[idx:].split()[1:]
            else:
                moveslist = []

            if params[1] == 'fen':
                if idx >= 0:
                    fenpart = smove[:idx]
                else:
                    fenpart = smove

                _, _, fen = fenpart.split(' ', 2)

            elif params[1] == 'startpos':
                fen = tools.FEN_INITIAL

            else:
                pass

            pos = tools.parseFEN(fen)
            color = WHITE if fen.split()[1] == 'w' else BLACK

            for move in moveslist:
                pos = pos.move(tools.mparse(color, move))
                color = 1 - color

        elif smove.startswith('go'):
            #  default options
            depth = 1000
            movetime = -1

            _, *params = smove.split(' ')
            for param, val in zip(*2*(iter(params),)):
                if param == 'depth':
                    depth = int(val)
                if param == 'movetime':
                    movetime = int(val)
                if param == 'wtime':
                    our_time = int(val)
                if param == 'btime':
                    opp_time = int(val)

            moves_remain = 40

            start = time.time()
            ponder = None
            for sdepth, _move, _score in searcher.search(pos):
                moves = tools.pv(searcher, pos, include_scores=False)

                if show_thinking:
                    entry = searcher.tp_score.get((pos, sdepth, True))
                    score = int(round((entry.lower + entry.upper)/2))
                    usedtime = int((time.time() - start) * 1000)
                    moves_str = moves if len(moves) < 15 else ''
                    output('info depth {} score cp {} time {} nodes {} pv {}'.format(sdepth, score, usedtime, searcher.nodes, moves_str))

                if len(moves) > 5:
                    ponder = moves[1]

                if movetime > 0 and (time.time() - start) * 1000 > movetime:
                    break

                if (time.time() - start) * 1000 > our_time/moves_remain:
                    break

                if sdepth >= depth:
                    break

            entry = searcher.tp_score.get((pos, sdepth, True))
            m, s = searcher.tp_move.get(pos), entry.lower
            # We only resign once we are mated.. That's never?
            if s == -sunfish.MATE_UPPER:
                output('resign')
            else:
                moves = moves.split(' ')
                if len(moves) > 1:
                    output(f'bestmove {moves[0]} ponder {moves[1]}')
                else:
                    output('bestmove ' + moves[0])

        elif smove.startswith('time'):
            our_time = int(smove.split()[1])

        elif smove.startswith('otim'):
            opp_time = int(smove.split()[1])

        else:
            pass

if __name__ == '__main__':
    main()



# Advanced UCI interface

import re, time
from concurrent.futures import ThreadPoolExecutor
from threading import Event
from functools import partial
from .. import sunfish

print = partial(print, flush=True)


def render_move(move, white_pov):
    if move is None:
        return "(none)"
    i, j = move.i, move.j
    if not white_pov:
        i, j = 119 - i, 119 - j
    render = sunfish.render
    return render(i) + render(j) + move.prom.lower()


def parse_move(move_str, white_pov):
    parse = sunfish.parse
    i, j, prom = (
        parse(move_str[:2]),
        parse(move_str[2:4]),
        move_str[4:].upper(),
    )
    if not white_pov:
        i, j = 119 - i, 119 - j
    return sunfish.Move(i, j, prom)


def go_loop(
    searcher, hist, stop_event, max_movetime=0, max_depth=0, debug=False
):
    if debug:
        print(f"Going movetime={max_movetime}, depth={max_depth}")

    start = time.time()
    best_move = None
    for depth, gamma, score, move in searcher.search(hist):
        # Our max_depth implementation is a bit wasteful.
        # We never know when we've seen the last at a certain depth
        # before we get to the next one
        if depth - 1 >= max_depth:
            break
        elapsed = time.time() - start
        fields = {
            "depth": depth,
            "time": round(1000 * elapsed),
            "nodes": searcher.nodes,
            "nps": round(searcher.nodes / elapsed),
        }
        if score >= gamma:
            fields["score cp"] = f"{score} lowerbound"
            best_move = render_move(move, white_pov=len(hist) % 2 == 1)
            fields["pv"] = " ".join(
                pv(searcher, hist[-1], include_scores=False)
            )
        else:
            fields["score cp"] = f"{score} upperbound"
        print("info", " ".join(f"{k} {v}" for k, v in fields.items()))

        # We may not have a move yet at depth = 1
        if depth > 1:
            if elapsed > max_movetime * 2 / 3:
                break
            if stop_event.is_set():
                break

    # FIXME: If we are in "go infinite" we aren't actually supposed to stop the
    # go-loop before we got stop_event. Unfortunately we currently don't know if
    # we are in "go infinite" since it's simply translated to "go depth 100".

    my_pv = pv(searcher, hist[-1], include_scores=False)
    print("bestmove", my_pv[0] if my_pv else "(none)")


def mate_loop(
    searcher,
    hist,
    stop_event,
    max_movetime=0,
    max_depth=0,
    find_draw=False,
    debug=False,
):
    start = time.time()
    for d in range(int(max_depth) + 1):
        if find_draw:
            s0 = searcher.bound(hist[-1], 0, d)
            elapsed = time.time() - start
            print("info", "depth", d, "score lowerbound cp", s0)
            s1 = searcher.bound(hist[-1], 1, d)
            elapsed = time.time() - start
            print("info", "depth", d, "score upperbound cp", s1)
            if s0 >= 0 and s1 < 1:
                break
        else:
            score = searcher.bound(hist[-1], sunfish.MATE_LOWER, d)
            elapsed = time.time() - start
            print(
                "info depth",
                d,
                "score cp",
                score,
                "time",
                round(1000 * elapsed),
                "pv",
                " ".join(pv(searcher, hist[-1], include_scores=False)),
            )
            if score >= sunfish.MATE_LOWER:
                break
        if elapsed > max_movetime:
            break
        if stop_event.is_set():
            break
    move = searcher.tp_move.get(hist[-1])
    move_str = render_move(move, white_pov=len(hist) % 2 == 1)
    print("bestmove", move_str)


def perft(pos, depth, debug=False):

    def _perft_count(pos, depth):
        # Check that we didn't get to an illegal position
        if can_kill_king(pos):
            return -1
        if depth == 0:
            return 1
        res = 0
        for move in pos.gen_moves():
            cnt = _perft_count(pos.move(move), depth - 1)
            if cnt != -1:
                res += cnt
        return res

    total = 0
    for move in pos.gen_moves():
        move_uci = render_move(move, get_color(pos) == WHITE)
        cnt = _perft_count(pos.move(move), depth - 1)
        if cnt != -1:
            print(f"{move_uci}: {cnt}")
            total += cnt
    print()
    print("Nodes searched:", total)


def run(sunfish_module, startpos):
    global sunfish
    sunfish = sunfish_module

    debug = False
    hist = [startpos]
    searcher = sunfish.Searcher()

    with ThreadPoolExecutor(max_workers=1) as executor:
        # Noop future to get started
        go_future = executor.submit(lambda: None)
        do_stop_event = Event()

        while True:
            try:
                args = input().split()
                if not args:
                    continue

                elif args[0] in ("stop", "quit"):
                    if go_future.running():
                        if debug:
                            print("Stopping go loop...")
                        do_stop_event.set()
                        go_future.result()
                    else:
                        if debug:
                            print("Go loop not running...")
                    if args[0] == "quit":
                        break

                elif not go_future.done():
                    print(f"Ignoring input {args}. Please call 'stop' first.")
                    continue

                # Make sure we are really done, and throw any errors that may have
                # happened in the go loop.
                go_future.result(timeout=0)

                if args[0] == "uci":
                    print(f"id name {sunfish.version}")
                    for attr, (lo, hi) in sunfish.opt_ranges.items():
                        default = getattr(sunfish, attr)
                        print(
                            f"option name {attr} type spin default {default} min {lo} max {hi}"
                        )
                    print("uciok")

                elif args[0] == "setoption":
                    _, uci_key, _, uci_value = args[1:]
                    setattr(sunfish, uci_key, int(uci_value))

                # FIXME: It seems we should reply to "isready" even while thinking.
                # See: https://talkchess.com/forum3/viewtopic.php?f=7&t=81233&start=10
                elif args[0] == "isready":
                    print("readyok")

                elif args[0] == "quit":
                    break

                elif args[:2] == ["position", "startpos"]:
                    hist = [startpos]
                    for ply, move in enumerate(args[3:]):
                        hist.append(
                            hist[-1].move(parse_move(move, ply % 2 == 0))
                        )

                elif args[:2] == ["position", "fen"]:
                    pos = from_fen(*args[2:8])
                    hist = (
                        [pos]
                        if get_color(pos) == WHITE
                        else [pos.rotate(), pos]
                    )
                    if len(args) > 8:
                        assert args[8] == "moves"
                        for move in args[9:]:
                            hist.append(
                                hist[-1].move(
                                    parse_move(move, len(hist) % 2 == 1)
                                )
                            )

                elif args[0] == "go":
                    think = 10**6
                    max_depth = 100
                    loop = go_loop

                    if args[1:] == [] or args[1] == "infinite":
                        pass

                    elif args[1] == "movetime":
                        movetime = args[2]
                        think = int(movetime) / 1000

                    elif args[1] == "wtime":
                        wtime, btime, winc, binc = [
                            int(a) / 1000 for a in args[2::2]
                        ]
                        # we always consider ourselves white, but uci doesn't
                        if len(hist) % 2 == 0:
                            wtime, winc = btime, binc
                        think = min(wtime / 40 + winc, wtime / 2 - 1)
                        # let's go fast for the first moves
                        if len(hist) < 3:
                            think = min(think, 1)

                    elif args[1] == "depth":
                        max_depth = int(args[2])

                    elif args[1] in ("mate", "draw"):
                        max_depth = int(args[2])
                        loop = partial(mate_loop, find_draw=args[1] == "draw")

                    elif args[1] == "perft":
                        perft(hist[-1], int(args[2]), debug=debug)
                        continue

                    do_stop_event.clear()
                    go_future = executor.submit(
                        loop,
                        searcher,
                        hist,
                        do_stop_event,
                        think,
                        max_depth,
                        debug=debug,
                    )

                    # Make sure we get informed if the job fails
                    def callback(fut):
                        fut.result(timeout=0)

                    go_future.add_done_callback(callback)

            except (KeyboardInterrupt, EOFError):
                if go_future.running():
                    if debug:
                        print("Stopping go loop...")
                    do_stop_event.set()
                    go_future.result()
                break


# Old tools stuff

WHITE, BLACK = range(2)


def from_fen(board, color, castling, enpas, _hclock, _fclock):
    board = re.sub(r"\d", (lambda m: "." * int(m.group(0))), board)
    board = list(21 * " " + "  ".join(board.split("/")) + 21 * " ")
    board[9::10] = ["\n"] * 12
    board = "".join(board)
    wc = ("Q" in castling, "K" in castling)
    bc = ("k" in castling, "q" in castling)
    ep = sunfish.parse(enpas) if enpas != "-" else 0
    if hasattr(sunfish, "features"):
        wf, bf = sunfish.features(board)
        pos = sunfish.Position(board, 0, wf, bf, wc, bc, ep, 0)
        pos = pos._replace(score=pos.calculate_score())
    else:
        score = sum(
            sunfish.pst[c][i] for i, c in enumerate(board) if c.isupper()
        )
        score -= sum(
            sunfish.pst[c.upper()][119 - i]
            for i, c in enumerate(board)
            if c.islower()
        )
        pos = sunfish.Position(board, score, wc, bc, ep, 0)
    return pos if color == "w" else pos.rotate()


def get_color(pos):
    """A slightly hacky way to to get the color from a sunfish position"""
    return BLACK if pos.board.startswith("\n") else WHITE


def can_kill_king(pos):
    # If we just checked for opponent moves capturing the king, we would miss
    # captures in case of illegal castling.
    # MATE_LOWER = 60_000 - 10 * 929
    # return any(pos.value(m) >= MATE_LOWER for m in pos.gen_moves())
    return any(
        pos.board[m.j] == "k" or abs(m.j - pos.kp) < 2 for m in pos.gen_moves()
    )


def pv(searcher, pos, include_scores=True, include_loop=False):
    res = []
    seen_pos = set()
    color = get_color(pos)
    origc = color
    if include_scores:
        res.append(str(pos.score))
    while True:
        if hasattr(pos, "wf"):
            move = searcher.tp_move.get(pos.hash())
        elif hasattr(searcher, "tp_move"):
            move = searcher.tp_move.get(pos)
        elif hasattr(searcher, "tt_new"):
            move = searcher.tt_new[0][pos, True].move
        # The tp may have illegal moves, given lower depths don't detect king killing
        if move is None or can_kill_king(pos.move(move)):
            break
        res.append(render_move(move, get_color(pos) == WHITE))
        pos, color = pos.move(move), 1 - color

        if hasattr(pos, "wf"):
            if pos.hash() in seen_pos:
                if include_loop:
                    res.append("loop")
                break
            seen_pos.add(pos.hash())
        else:
            if pos in seen_pos:
                if include_loop:
                    res.append("loop")
                break
            seen_pos.add(pos)

        if include_scores:
            res.append(str(pos.score if color == origc else -pos.score))
    return res



import tools.uci
import chess
import chess.engine

board = chess.Board()
print(board)

engine = chess.engine.SimpleEngine.popen_uci(
    r"/users/meni/code/personal/chess/sunfish/build/pack.sh"
)

while not board.is_game_over():
    result = engine.play(board, chess.engine.Limit(time=0.1))
    board.push(result.move)

engine.quit()



import sys
import chess
import argparse
from movegeneration import next_move


def talk():
    """
    The main input/output loop.
    This implements a slice of the UCI protocol.
    """
    board = chess.Board()
    depth = get_depth()

    while True:
        msg = input()
        command(depth, board, msg)


def command(depth: int, board: chess.Board, msg: str):
    """
    Accept UCI commands and respond.
    The board state is also updated.
    """
    msg = msg.strip()
    tokens = msg.split(" ")
    while "" in tokens:
        tokens.remove("")

    if msg == "quit":
        sys.exit()

    if msg == "uci":
        print("id name Andoma")  # Andrew/Roma -> And/oma
        print("id author Andrew Healey & Roma Parramore")
        print("uciok")
        return

    if msg == "isready":
        print("readyok")
        return

    if msg == "ucinewgame":
        return

    if msg.startswith("position"):
        if len(tokens) < 2:
            return

        # Set starting position
        if tokens[1] == "startpos":
            board.reset()
            moves_start = 2
        elif tokens[1] == "fen":
            fen = " ".join(tokens[2:8])
            board.set_fen(fen)
            moves_start = 8
        else:
            return

        # Apply moves
        if len(tokens) <= moves_start or tokens[moves_start] != "moves":
            return

        for move in tokens[(moves_start+1):]:
            board.push_uci(move)

    if msg == "d":
        # Non-standard command, but supported by Stockfish and helps debugging
        print(board)
        print(board.fen())

    if msg[0:2] == "go":
        _move = next_move(depth, board)
        print(f"bestmove {_move}")
        return


def get_depth() -> int:
    parser = argparse.ArgumentParser()
    parser.add_argument("--depth", default=3, help="provide an integer (default: 3)")
    args = parser.parse_args()
    return max([1, int(args.depth)])



#!/usr/bin/env pypy3
from __future__ import print_function

import time, math
from itertools import count
from collections import namedtuple, defaultdict

# If we could rely on the env -S argument, we could just use "pypy3 -u"
# as the shebang to unbuffer stdout. But alas we have to do this instead:
# from functools import partial
# print = partial(print, flush=True)

version = "sunfish 2023"

###############################################################################
# Piece-Square tables. Tune these to change sunfish's behaviour
###############################################################################

# With xz compression this whole section takes 652 bytes.
# That's pretty good given we have 64*6 = 384 values.
# Though probably we could do better...
# For one thing, they could easily all fit into int8.
piece = {"P": 100, "N": 280, "B": 320, "R": 479, "Q": 929, "K": 60000}
pst = {
    "P": (
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        78,
        83,
        86,
        73,
        102,
        82,
        85,
        90,
        7,
        29,
        21,
        44,
        40,
        31,
        44,
        7,
        -17,
        16,
        -2,
        15,
        14,
        0,
        15,
        -13,
        -26,
        3,
        10,
        9,
        6,
        1,
        0,
        -23,
        -22,
        9,
        5,
        -11,
        -10,
        -2,
        3,
        -19,
        -31,
        8,
        -7,
        -37,
        -36,
        -14,
        3,
        -31,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
    ),
    "N": (
        -66,
        -53,
        -75,
        -75,
        -10,
        -55,
        -58,
        -70,
        -3,
        -6,
        100,
        -36,
        4,
        62,
        -4,
        -14,
        10,
        67,
        1,
        74,
        73,
        27,
        62,
        -2,
        24,
        24,
        45,
        37,
        33,
        41,
        25,
        17,
        -1,
        5,
        31,
        21,
        22,
        35,
        2,
        0,
        -18,
        10,
        13,
        22,
        18,
        15,
        11,
        -14,
        -23,
        -15,
        2,
        0,
        2,
        0,
        -23,
        -20,
        -74,
        -23,
        -26,
        -24,
        -19,
        -35,
        -22,
        -69,
    ),
    "B": (
        -59,
        -78,
        -82,
        -76,
        -23,
        -107,
        -37,
        -50,
        -11,
        20,
        35,
        -42,
        -39,
        31,
        2,
        -22,
        -9,
        39,
        -32,
        41,
        52,
        -10,
        28,
        -14,
        25,
        17,
        20,
        34,
        26,
        25,
        15,
        10,
        13,
        10,
        17,
        23,
        17,
        16,
        0,
        7,
        14,
        25,
        24,
        15,
        8,
        25,
        20,
        15,
        19,
        20,
        11,
        6,
        7,
        6,
        20,
        16,
        -7,
        2,
        -15,
        -12,
        -14,
        -15,
        -10,
        -10,
    ),
    "R": (
        35,
        29,
        33,
        4,
        37,
        33,
        56,
        50,
        55,
        29,
        56,
        67,
        55,
        62,
        34,
        60,
        19,
        35,
        28,
        33,
        45,
        27,
        25,
        15,
        0,
        5,
        16,
        13,
        18,
        -4,
        -9,
        -6,
        -28,
        -35,
        -16,
        -21,
        -13,
        -29,
        -46,
        -30,
        -42,
        -28,
        -42,
        -25,
        -25,
        -35,
        -26,
        -46,
        -53,
        -38,
        -31,
        -26,
        -29,
        -43,
        -44,
        -53,
        -30,
        -24,
        -18,
        5,
        -2,
        -18,
        -31,
        -32,
    ),
    "Q": (
        6,
        1,
        -8,
        -104,
        69,
        24,
        88,
        26,
        14,
        32,
        60,
        -10,
        20,
        76,
        57,
        24,
        -2,
        43,
        32,
        60,
        72,
        63,
        43,
        2,
        1,
        -16,
        22,
        17,
        25,
        20,
        -13,
        -6,
        -14,
        -15,
        -2,
        -5,
        -1,
        -10,
        -20,
        -22,
        -30,
        -6,
        -13,
        -11,
        -16,
        -11,
        -16,
        -27,
        -36,
        -18,
        0,
        -19,
        -15,
        -15,
        -21,
        -38,
        -39,
        -30,
        -31,
        -13,
        -31,
        -36,
        -34,
        -42,
    ),
    "K": (
        4,
        54,
        47,
        -99,
        -99,
        60,
        83,
        -62,
        -32,
        10,
        55,
        56,
        56,
        55,
        10,
        3,
        -62,
        12,
        -57,
        44,
        -67,
        28,
        37,
        -31,
        -55,
        50,
        11,
        -4,
        -19,
        13,
        0,
        -49,
        -55,
        -43,
        -52,
        -28,
        -51,
        -47,
        -8,
        -50,
        -47,
        -42,
        -43,
        -79,
        -64,
        -32,
        -29,
        -32,
        -4,
        3,
        -14,
        -50,
        -57,
        -18,
        13,
        4,
        17,
        30,
        -3,
        -14,
        6,
        -1,
        40,
        18,
    ),
}
# Pad tables and join piece and pst dictionaries
for k, table in pst.items():
    padrow = lambda row: (0,) + tuple(x + piece[k] for x in row) + (0,)
    pst[k] = sum((padrow(table[i * 8 : i * 8 + 8]) for i in range(8)), ())
    pst[k] = (0,) * 20 + pst[k] + (0,) * 20

###############################################################################
# Global constants
###############################################################################

# Our board is represented as a 120 character string. The padding allows for
# fast detection of moves that don't stay within the board.
A1, H1, A8, H8 = 91, 98, 21, 28
initial = (
    "         \n"  #   0 -  9
    "         \n"  #  10 - 19
    " rnbqkbnr\n"  #  20 - 29
    " pppppppp\n"  #  30 - 39
    " ........\n"  #  40 - 49
    " ........\n"  #  50 - 59
    " ........\n"  #  60 - 69
    " ........\n"  #  70 - 79
    " PPPPPPPP\n"  #  80 - 89
    " RNBQKBNR\n"  #  90 - 99
    "         \n"  # 100 -109
    "         \n"  # 110 -119
)

# Lists of possible moves for each piece type.
N, E, S, W = -10, 1, 10, -1
directions = {
    "P": (N, N + N, N + W, N + E),
    "N": (
        N + N + E,
        E + N + E,
        E + S + E,
        S + S + E,
        S + S + W,
        W + S + W,
        W + N + W,
        N + N + W,
    ),
    "B": (N + E, S + E, S + W, N + W),
    "R": (N, E, S, W),
    "Q": (N, E, S, W, N + E, S + E, S + W, N + W),
    "K": (N, E, S, W, N + E, S + E, S + W, N + W),
}

# Mate value must be greater than 8*queen + 2*(rook+knight+bishop)
# King value is set to twice this value such that if the opponent is
# 8 queens up, but we got the king, we still exceed MATE_VALUE.
# When a MATE is detected, we'll set the score to MATE_UPPER - plies to get there
# E.g. Mate in 3 will be MATE_UPPER - 6
MATE_LOWER = piece["K"] - 10 * piece["Q"]
MATE_UPPER = piece["K"] + 10 * piece["Q"]

# Constants for tuning search
QS = 40
QS_A = 140
EVAL_ROUGHNESS = 15

# minifier-hide start
opt_ranges = dict(
    QS=(0, 300),
    QS_A=(0, 300),
    EVAL_ROUGHNESS=(0, 50),
)
# minifier-hide end


###############################################################################
# Chess logic
###############################################################################


Move = namedtuple("Move", "i j prom")


class Position(namedtuple("Position", "board score wc bc ep kp")):
    """A state of a chess game
    board -- a 120 char representation of the board
    score -- the board evaluation
    wc -- the castling rights, [west/queen side, east/king side]
    bc -- the opponent castling rights, [west/king side, east/queen side]
    ep - the en passant square
    kp - the king passant square
    """

    def gen_moves(self):
        # For each of our pieces, iterate through each possible 'ray' of moves,
        # as defined in the 'directions' map. The rays are broken e.g. by
        # captures or immediately in case of pieces such as knights.
        for i, p in enumerate(self.board):
            if not p.isupper():
                continue
            for d in directions[p]:
                for j in count(i + d, d):
                    q = self.board[j]
                    # Stay inside the board, and off friendly pieces
                    if q.isspace() or q.isupper():
                        break
                    # Pawn move, double move and capture
                    if p == "P":
                        if d in (N, N + N) and q != ".":
                            break
                        if d == N + N and (
                            i < A1 + N or self.board[i + N] != "."
                        ):
                            break
                        if (
                            d in (N + W, N + E)
                            and q == "."
                            and j
                            not in (self.ep, self.kp, self.kp - 1, self.kp + 1)
                            # and j != self.ep and abs(j - self.kp) >= 2
                        ):
                            break
                        # If we move to the last row, we can be anything
                        if A8 <= j <= H8:
                            for prom in "NBRQ":
                                yield Move(i, j, prom)
                            break
                    # Move it
                    yield Move(i, j, "")
                    # Stop crawlers from sliding, and sliding after captures
                    if p in "PNK" or q.islower():
                        break
                    # Castling, by sliding the rook next to the king
                    if i == A1 and self.board[j + E] == "K" and self.wc[0]:
                        yield Move(j + E, j + W, "")
                    if i == H1 and self.board[j + W] == "K" and self.wc[1]:
                        yield Move(j + W, j + E, "")

    def rotate(self, nullmove=False):
        """Rotates the board, preserving enpassant, unless nullmove"""
        return Position(
            self.board[::-1].swapcase(),
            -self.score,
            self.bc,
            self.wc,
            119 - self.ep if self.ep and not nullmove else 0,
            119 - self.kp if self.kp and not nullmove else 0,
        )

    def move(self, move):
        i, j, prom = move
        p, q = self.board[i], self.board[j]
        put = lambda board, i, p: board[:i] + p + board[i + 1 :]
        # Copy variables and reset ep and kp
        board = self.board
        wc, bc, ep, kp = self.wc, self.bc, 0, 0
        score = self.score + self.value(move)
        # Actual move
        board = put(board, j, board[i])
        board = put(board, i, ".")
        # Castling rights, we move the rook or capture the opponent's
        if i == A1:
            wc = (False, wc[1])
        if i == H1:
            wc = (wc[0], False)
        if j == A8:
            bc = (bc[0], False)
        if j == H8:
            bc = (False, bc[1])
        # Castling
        if p == "K":
            wc = (False, False)
            if abs(j - i) == 2:
                kp = (i + j) // 2
                board = put(board, A1 if j < i else H1, ".")
                board = put(board, kp, "R")
        # Pawn promotion, double move and en passant capture
        if p == "P":
            if A8 <= j <= H8:
                board = put(board, j, prom)
            if j - i == 2 * N:
                ep = i + N
            if j == self.ep:
                board = put(board, j + S, ".")
        # We rotate the returned position, so it's ready for the next player
        return Position(board, score, wc, bc, ep, kp).rotate()

    def value(self, move):
        i, j, prom = move
        p, q = self.board[i], self.board[j]
        # Actual move
        score = pst[p][j] - pst[p][i]
        # Capture
        if q.islower():
            score += pst[q.upper()][119 - j]
        # Castling check detection
        if abs(j - self.kp) < 2:
            score += pst["K"][119 - j]
        # Castling
        if p == "K" and abs(i - j) == 2:
            score += pst["R"][(i + j) // 2]
            score -= pst["R"][A1 if j < i else H1]
        # Special pawn stuff
        if p == "P":
            if A8 <= j <= H8:
                score += pst[prom][j] - pst["P"][j]
            if j == self.ep:
                score += pst["P"][119 - (j + S)]
        return score


###############################################################################
# Search logic
###############################################################################

# lower <= s(pos) <= upper
Entry = namedtuple("Entry", "lower upper")


class Searcher:
    def __init__(self):
        self.tp_score = {}
        self.tp_move = {}
        self.history = set()
        self.nodes = 0

    def bound(self, pos, gamma, depth, can_null=True):
        """Let s* be the "true" score of the sub-tree we are searching.
        The method returns r, where
        if gamma >  s* then s* <= r < gamma  (A better upper bound)
        if gamma <= s* then gamma <= r <= s* (A better lower bound)"""
        self.nodes += 1

        # Depth <= 0 is QSearch. Here any position is searched as deeply as is needed for
        # calmness, and from this point on there is no difference in behaviour depending on
        # depth, so so there is no reason to keep different depths in the transposition table.
        depth = max(depth, 0)

        # Sunfish is a king-capture engine, so we should always check if we
        # still have a king. Notice since this is the only termination check,
        # the remaining code has to be comfortable with being mated, stalemated
        # or able to capture the opponent king.
        if pos.score <= -MATE_LOWER:
            return -MATE_UPPER

        # Look in the table if we have already searched this position before.
        # We also need to be sure, that the stored search was over the same
        # nodes as the current search.
        entry = self.tp_score.get(
            (pos, depth, can_null), Entry(-MATE_UPPER, MATE_UPPER)
        )
        if entry.lower >= gamma:
            return entry.lower
        if entry.upper < gamma:
            return entry.upper

        # Let's not repeat positions. We don't chat
        # - at the root (can_null=False) since it is in history, but not a draw.
        # - at depth=0, since it would be expensive and break "futulity pruning".
        if can_null and depth > 0 and pos in self.history:
            return 0

        # Generator of moves to search in order.
        # This allows us to define the moves, but only calculate them if needed.
        def moves():
            # First try not moving at all. We only do this if there is at least one major
            # piece left on the board, since otherwise zugzwangs are too dangerous.
            # FIXME: We also can't null move if we can capture the opponent king.
            # Since if we do, we won't spot illegal moves that could lead to stalemate.
            # For now we just solve this by not using null-move in very unbalanced positions.
            # TODO: We could actually use null-move in QS as well. Not sure it would be very useful.
            # But still.... We just have to move stand-pat to be before null-move.
            # if depth > 2 and can_null and any(c in pos.board for c in "RBNQ"):
            # if depth > 2 and can_null and any(c in pos.board for c in "RBNQ") and abs(pos.score) < 500:
            if depth > 2 and can_null and abs(pos.score) < 500:
                yield None, -self.bound(
                    pos.rotate(nullmove=True), 1 - gamma, depth - 3
                )

            # For QSearch we have a different kind of null-move, namely we can just stop
            # and not capture anything else.
            if depth == 0:
                yield None, pos.score

            # Look for the strongest ove from last time, the hash-move.
            killer = self.tp_move.get(pos)

            # If there isn't one, try to find one with a more shallow search.
            # This is known as Internal Iterative Deepening (IID). We set
            # can_null=True, since we want to make sure we actually find a move.
            if not killer and depth > 2:
                self.bound(pos, gamma, depth - 3, can_null=False)
                killer = self.tp_move.get(pos)

            # If depth == 0 we only try moves with high intrinsic score (captures and
            # promotions). Otherwise we do all moves. This is called quiescent search.
            val_lower = QS - depth * QS_A

            # Only play the move if it would be included at the current val-limit,
            # since otherwise we'd get search instability.
            # We will search it again in the main loop below, but the tp will fix
            # things for us.
            if killer and pos.value(killer) >= val_lower:
                yield killer, -self.bound(
                    pos.move(killer), 1 - gamma, depth - 1
                )

            # Then all the other moves
            for val, move in sorted(
                ((pos.value(m), m) for m in pos.gen_moves()), reverse=True
            ):
                # Quiescent search
                if val < val_lower:
                    break

                # If the new score is less than gamma, the opponent will for sure just
                # stand pat, since ""pos.score + val < gamma === -(pos.score + val) >= 1-gamma""
                # This is known as futility pruning.
                if depth <= 1 and pos.score + val < gamma:
                    # Need special case for MATE, since it would normally be caught
                    # before standing pat.
                    yield move, (
                        pos.score + val if val < MATE_LOWER else MATE_UPPER
                    )
                    # We can also break, since we have ordered the moves by value,
                    # so it can't get any better than this.
                    break

                yield move, -self.bound(pos.move(move), 1 - gamma, depth - 1)

        # Run through the moves, shortcutting when possible
        best = -MATE_UPPER
        for move, score in moves():
            best = max(best, score)
            if best >= gamma:
                # Save the move for pv construction and killer heuristic
                if move is not None:
                    self.tp_move[pos] = move
                break

        # Stalemate checking is a bit tricky: Say we failed low, because
        # we can't (legally) move and so the (real) score is -infty.
        # At the next depth we are allowed to just return r, -infty <= r < gamma,
        # which is normally fine.
        # However, what if gamma = -10 and we don't have any legal moves?
        # Then the score is actaully a draw and we should fail high!
        # Thus, if best < gamma and best < 0 we need to double check what we are doing.

        # We will fix this problem another way: We add the requirement to bound, that
        # it always returns MATE_UPPER if the king is capturable. Even if another move
        # was also sufficient to go above gamma. If we see this value we know we are either
        # mate, or stalemate. It then suffices to check whether we're in check.

        # Note that at low depths, this may not actually be true, since maybe we just pruned
        # all the legal moves. So sunfish may report "mate", but then after more search
        # realize it's not a mate after all. That's fair.

        # This is too expensive to test at depth == 0
        if depth > 2 and best == -MATE_UPPER:
            flipped = pos.rotate(nullmove=True)
            # Hopefully this is already in the TT because of null-move
            in_check = self.bound(flipped, MATE_UPPER, 0) == MATE_UPPER
            best = -MATE_LOWER if in_check else 0

        # Table part 2
        if best >= gamma:
            self.tp_score[pos, depth, can_null] = Entry(best, entry.upper)
        if best < gamma:
            self.tp_score[pos, depth, can_null] = Entry(entry.lower, best)

        return best

    def search(self, history):
        """Iterative deepening MTD-bi search"""
        self.nodes = 0
        self.history = set(history)
        self.tp_score.clear()

        gamma = 0
        # In finished games, we could potentially go far enough to cause a recursion
        # limit exception. Hence we bound the ply. We also can't start at 0, since
        # that's quiscent search, and we don't always play legal moves there.
        for depth in range(1, 1000):
            # The inner loop is a binary search on the score of the position.
            # Inv: lower <= score <= upper
            # 'while lower != upper' would work, but it's too much effort to spend
            # on what's probably not going to change the move played.
            lower, upper = -MATE_LOWER, MATE_LOWER
            while lower < upper - EVAL_ROUGHNESS:
                score = self.bound(history[-1], gamma, depth, can_null=False)
                if score >= gamma:
                    lower = score
                if score < gamma:
                    upper = score
                yield depth, gamma, score, self.tp_move.get(history[-1])
                gamma = (lower + upper + 1) // 2


###############################################################################
# UCI User interface
###############################################################################


def parse(c):
    fil, rank = ord(c[0]) - ord("a"), int(c[1]) - 1
    return A1 + fil - 10 * rank


def render(i):
    rank, fil = divmod(i - A1, 10)
    return chr(fil + ord("a")) + str(-rank + 1)


# hist = [Position(initial, 0, (True, True), (True, True), 0, 0)]

# input = raw_input

# minifier-hide start
# import sys, tools.uci

# tools.uci.run(sys.modules[__name__], hist[-1])
# sys.exit()
# minifier-hide end


# def run():
#     searcher = Searcher()
#     while True:
#         args = input().split()
#         if args[0] == "uci":
#             print("id name", version)
#             print("uciok")
#
#         elif args[0] == "isready":
#             print("readyok")
#
#         elif args[0] == "quit":
#             break
#
#         elif args[:2] == ["position", "startpos"]:
#             del hist[1:]
#             for ply, move in enumerate(args[3:]):
#                 i, j, prom = (
#                     parse(move[:2]),
#                     parse(move[2:4]),
#                     move[4:].upper(),
#                 )
#                 if ply % 2 == 1:
#                     i, j = 119 - i, 119 - j
#                 hist.append(hist[-1].move(Move(i, j, prom)))
#
#         elif args[0] == "go":
#             wtime, btime, winc, binc = [int(a) / 1000 for a in args[2::2]]
#             if len(hist) % 2 == 0:
#                 wtime, winc = btime, binc
#             think = min(wtime / 40 + winc, wtime / 2 - 1)
#
#             start = time.time()
#             move_str = None
#             for depth, gamma, score, move in Searcher().search(hist):
#                 # The only way we can be sure to have the real move in tp_move,
#                 # is if we have just failed high.
#                 if score >= gamma:
#                     i, j = move.i, move.j
#                     if len(hist) % 2 == 0:
#                         i, j = 119 - i, 119 - j
#                     move_str = render(i) + render(j) + move.prom.lower()
#                     print(
#                         "info depth", depth, "score cp", score, "pv", move_str
#                     )
#                 if move_str and time.time() - start > think * 0.8:
#                     break
#
#             print("bestmove", move_str or "(none)")
#
#
# if __name__ == "__main__":
#     run()



import time
from .sunfish import (
    Searcher,
    Position,
    initial,
    parse,
    render,
    Move,
)


def uci_to_sun(move, ply):
    i, j, prom = (
        parse(move[:2]),
        parse(move[2:4]),
        move[4:].upper(),
    )
    if ply % 2 == 1:
        i, j = 119 - i, 119 - j

    return Move(i, j, prom)


def sun_to_uci(move, hist):
    i, j = move.i, move.j
    if len(hist) % 2 == 0:
        i, j = 119 - i, 119 - j
    move_str = render(i) + render(j) + move.prom.lower()
    return move_str


def generate_move(hist, bwtime, bwinc=0):
    wtime, btime, winc, binc = bwtime, bwtime, bwinc, bwinc

    if len(hist) % 2 == 0:
        wtime, winc = btime, binc

    think = min(wtime / 40 + winc, wtime / 2 - 1)

    start = time.time()
    move_str = None
    move = None
    for depth, gamma, score, move in Searcher().search(hist):
        if score >= gamma:
            move_str = sun_to_uci(move, hist)
        if move_str and time.time() - start > think * 0.8:
            break

    return move, move_str


def starting_hist():
    return [Position(initial, 0, (True, True), (True, True), 0, 0)]



# inspired by the https://github.com/thomasahle/sunfish user inferface

import chess
import argparse
from movegeneration import next_move


def start():
    """
    Start the command line user interface.
    """
    board = chess.Board()
    user_side = (
        chess.WHITE if input("Start as [w]hite or [b]lack:\n") == "w" else chess.BLACK
    )

    if user_side == chess.WHITE:
        print(render(board))
        board.push(get_move(board))

    while not board.is_game_over():
        board.push(next_move(get_depth(), board, debug=False))
        print(render(board))
        board.push(get_move(board))

    print(f"\nResult: [w] {board.result()} [b]")


def render(board: chess.Board) -> str:
    """
    Print a side-relative chess board with special chess characters.
    """
    board_string = list(str(board))
    uni_pieces = {
        "R": "♖",
        "N": "♘",
        "B": "♗",
        "Q": "♕",
        "K": "♔",
        "P": "♙",
        "r": "♜",
        "n": "♞",
        "b": "♝",
        "q": "♛",
        "k": "♚",
        "p": "♟",
        ".": "·",
    }
    for idx, char in enumerate(board_string):
        if char in uni_pieces:
            board_string[idx] = uni_pieces[char]
    ranks = ["1", "2", "3", "4", "5", "6", "7", "8"]
    display = []
    for rank in "".join(board_string).split("\n"):
        display.append(f"  {ranks.pop()} {rank}")
    if board.turn == chess.BLACK:
        display.reverse()
    display.append("    a b c d e f g h")
    return "\n" + "\n".join(display)


def get_move(board: chess.Board) -> chess.Move:
    """
    Try (and keep trying) to get a legal next move from the user.
    Play the move by mutating the game board.
    """
    move = input(f"\nYour move (e.g. {list(board.legal_moves)[0]}):\n")

    for legal_move in board.legal_moves:
        if move == str(legal_move):
            return legal_move
    return get_move(board)


def get_depth() -> int:
    parser = argparse.ArgumentParser()
    parser.add_argument("--depth", default=3, help="provide an integer (default: 3)")
    args = parser.parse_args()
    return max([1, int(args.depth)])


if __name__ == "__main__":
    try:
        start()
    except KeyboardInterrupt:
        pass



#!/usr/bin/env -S uv run
import argparse
import random
import subprocess
import io
import os
import shlex
from datetime import date, datetime

import chess
import chess.pgn
import chess.polyglot
import chess.engine

from .engines.andoma.movegeneration import next_move as andoma_gen
from .engines.sunfish import sunfish_uci
from .engines.sunfish.tools import uci


# --- Colors (ANSI) ------------------------------------------------------------
# Works in most terminals. On Windows, ANSI is supported in modern terminals;
# if you need broader support, install `colorama` and it will be used if present.
try:
    import colorama  # type: ignore
    colorama.just_fix_windows_console()
except Exception:
    pass


class Style:
    RESET = "\033[0m"
    BOLD = "\033[1m"
    DIM = "\033[2m"

    RED = "\033[31m"
    GREEN = "\033[32m"
    YELLOW = "\033[33m"
    BLUE = "\033[34m"
    MAGENTA = "\033[35m"
    CYAN = "\033[36m"


def c(text: str, *styles: str) -> str:
    return "".join(styles) + text + Style.RESET


# --- Game logic --------------------------------------------------------------
class Game:
    def __init__(
        self,
        engine_kind: str,
        engine_name: str,
        player_color: chess.Color,
        engine: chess.engine.SimpleEngine | None = None,
        verbose: bool = False,
        book_path: str | None = None,
        book_chance: float = 0.5,
    ):
        self.board = chess.Board()
        self.engine_kind = engine_kind  # "random", "andoma", "sunfish", "uci"
        self.engine_name = engine_name
        self.player_color = player_color
        self.engine = engine
        self.verbose = verbose
        self.book_path = book_path
        self.book_chance = book_chance
        self.turn = chess.WHITE  # whose turn it is to move in our bookkeeping
        self.count = 0           # move number (full moves)
        self.pgn_text = ""
        self.ended = False

    def reset(self):
        self.board.set_fen(chess.STARTING_FEN)
        self.turn = chess.WHITE
        self.count = 0
        self.pgn_text = ""
        self.ended = False

    def close_engine(self):
        if self.engine is None:
            return
        try:
            self.engine.quit()
        except Exception:
            try:
                self.engine.close()
            except Exception:
                pass
        self.engine = None


def is_a_draw(board: chess.Board):
    """Return (is_draw, message)."""
    if board.is_stalemate():
        return True, "Stalemate"
    elif board.is_insufficient_material():
        return True, "Insufficient Material"
    elif board.can_claim_fifty_moves():
        return True, "Fifty-move rule"
    elif board.can_claim_threefold_repetition():
        return True, "Threefold repetition"
    return False, ""


def bool_color_to_string(color_b: chess.Color) -> str:
    return "white" if color_b == chess.WHITE else "black"


def finalize_pgn(pgn_str: str, player_color: chess.Color, engine_name: str):
    final_pgn = pgn_str + "\n\n"
    game_pgn = chess.pgn.read_game(io.StringIO(final_pgn))
    game_pgn.headers["Event"] = "Blind-chess match"
    game_pgn.headers["Site"] = "Terminal"
    if player_color == chess.WHITE:
        game_pgn.headers["White"] = "Me"
        game_pgn.headers["Black"] = f"{engine_name} Bot"
    else:
        game_pgn.headers["White"] = f"{engine_name} Bot"
        game_pgn.headers["Black"] = "Me"
    game_pgn.headers["Date"] = date.today().isoformat()
    return game_pgn


def bot_makes_a_move(game: Game):
    board = game.board
    move = random.choice(list(board.legal_moves))
    used_book = False

    if game.engine_kind == "andoma":
        move = andoma_gen(depth=4, board=board, debug=False)
    elif game.engine_kind == "sunfish":
        position = uci.from_fen(*board.fen().split())
        current_hist = (
            [position]
            if uci.get_color(position) == uci.WHITE
            else [position.rotate(), position]
        )
        total_time = random.randint(10, 60)
        _, uci_move_str = sunfish_uci.generate_move(current_hist, total_time)
        move = chess.Move.from_uci(uci_move_str)
    elif game.engine_kind == "uci":
        if game.engine is None:
            raise RuntimeError("UCI engine is not initialized.")
        think_time = random.uniform(0.1, 0.5)
        result = game.engine.play(board, chess.engine.Limit(time=think_time))
        move = result.move

    # optional opening book
    roll = random.random()
    if game.book_path and game.count < 15 and roll < game.book_chance:
        try:
            with chess.polyglot.open_reader(game.book_path) as reader:
                move = reader.weighted_choice(board).move
                used_book = True
        except (IndexError, FileNotFoundError):
            pass

    turn_number = board.fullmove_number
    move_san = board.san(move)
    board.push(move)

    if game.turn == chess.WHITE:
        game.count += 1
        game.pgn_text += f"\n{game.count}. {move_san}"
    else:
        game.pgn_text += f" {move_san}"

    # Minimal engine output (colored, no label)
    turn_prefix = c(f"{turn_number}.", Style.DIM)
    if game.verbose and used_book:
        print(
            f"{turn_prefix} {c(move_san, Style.MAGENTA, Style.BOLD)}"
            f"{c(' (book)', Style.DIM)}"
        )
    else:
        print(f"{turn_prefix} {c(move_san, Style.MAGENTA, Style.BOLD)}")

    # draw / checkmate handling
    check_draw, draw_type = is_a_draw(board)
    if check_draw:
        print(c(f"Draw: {draw_type}", Style.YELLOW, Style.BOLD))
        game.pgn_text += " { The game is a draw. } 1/2-1/2"
        game.ended = True
        return

    if board.is_checkmate():
        print(c("Checkmate.", Style.RED, Style.BOLD))
        result = "0-1" if game.player_color == chess.WHITE else "1-0"
        game.pgn_text += (
            f" {{ {bool_color_to_string(not game.player_color)} wins by checkmate. }} "
            f"{result}"
        )
        game.ended = True
        return

    game.turn = not game.turn


def _engine_display_name(command: str, engine: chess.engine.SimpleEngine) -> str:
    name = engine.id.get("name")
    if name:
        return name
    return os.path.basename(command.split()[0]) or "UCI"


def _spawn_uci_engine(command: str) -> chess.engine.SimpleEngine | None:
    cmd = shlex.split(command)
    if not cmd:
        return None
    try:
        return chess.engine.SimpleEngine.popen_uci(cmd, stderr=subprocess.DEVNULL)
    except (FileNotFoundError, PermissionError, OSError, chess.engine.EngineError):
        return None


def _book_status_line(book_path: str | None, book_chance: float) -> str:
    if book_path is None:
        return "Opening book: none"
    if os.path.exists(book_path):
        chance_pct = round(book_chance * 100)
        return f"Opening book: {book_path} ({chance_pct}%)"
    return "Opening book: none"


def choose_engine() -> tuple[str, str, chess.engine.SimpleEngine | None]:
    options = ["random", "andoma", "sunfish", "uci"]
    print(c("Choose engine:", Style.CYAN, Style.BOLD))
    for i, name in enumerate(options, start=1):
        print(f"  {c(str(i) + '.', Style.DIM)} {c(name, Style.CYAN)}")
    while True:
        choice = input(c("engine> ", Style.DIM)).strip().lower()
        if choice in {"1", "2", "3", "4"}:
            choice = options[int(choice) - 1]
        if choice in options:
            if choice != "uci":
                return choice, choice, None
            while True:
                cmd = input(c("uci engine path/command> ", Style.DIM)).strip()
                engine = _spawn_uci_engine(cmd)
                if engine is not None:
                    display_name = _engine_display_name(cmd, engine)
                    return "uci", display_name, engine
                print(c("Could not start engine. Try again.", Style.RED))
        print(c("Invalid choice.", Style.RED))


def choose_color():
    print(c("Choose your color:", Style.CYAN, Style.BOLD))
    print(f"  {c('1.', Style.DIM)} {c('white', Style.CYAN)}")
    print(f"  {c('2.', Style.DIM)} {c('black', Style.CYAN)}")
    print(f"  {c('3.', Style.DIM)} {c('random', Style.CYAN)}")
    while True:
        choice = input(c("color> ", Style.DIM)).strip().lower()
        if choice in {"1", "white"}:
            return chess.WHITE
        if choice in {"2", "black"}:
            return chess.BLACK
        if choice in {"3", "random"}:
            return random.choice([chess.WHITE, chess.BLACK])
        print(c("Invalid choice.", Style.RED))


def print_help():
    print(c("Lobby:", Style.CYAN, Style.BOLD))
    print(f"  {c('start', Style.CYAN)}  start a new game")
    print(f"  {c('quick', Style.CYAN)}  start with sunfish + random color")
    print(f"  {c('help', Style.CYAN)}   show this help")
    print(f"  {c('quit', Style.CYAN)}   quit")
    print()
    print(c("In-game:", Style.CYAN, Style.BOLD))
    print(f"  {c('show', Style.CYAN)}   show the board")
    print(f"  {c('moves', Style.CYAN)}  show legal moves (SAN)")
    print(f"  {c('fen', Style.CYAN)}    show FEN")
    print(f"  {c('pgn', Style.CYAN)}    show PGN so far")
    print(f"  {c('resign', Style.CYAN)} resign the game")
    print()
    print(c("Or type a move in SAN, e.g. e4, Nf3, exd5, a8=Q.", Style.DIM))


def parse_command(s: str):
    """Normalize commands: ':show' and 'show' both become 'show'."""
    s = s.strip()
    if not s:
        return None
    if s.startswith(":"):
        s = s[1:]
    return s.lower()


def ask_yes_no(prompt: str, default_no: bool = True) -> bool:
    suffix = " [y/N]: " if default_no else " [Y/n]: "
    while True:
        ans = input(c(prompt + suffix, Style.DIM)).strip().lower()
        if not ans:
            return not default_no
        if ans in {"y", "yes"}:
            return True
        if ans in {"n", "no"}:
            return False
        print(c("Please answer y or n.", Style.RED))


def _slugify_filename(text: str) -> str:
    cleaned = []
    for ch in text.strip().lower():
        if ch.isalnum() or ch in {"-", "_"}:
            cleaned.append(ch)
        elif ch.isspace():
            cleaned.append("_")
    return "".join(cleaned) or "bot"


def ask_pgn_action() -> str:
    prompt = "Final PGN: (p)rint, (s)ave, or (n)one?"
    suffix = " [n]: "
    while True:
        ans = input(c(prompt + suffix, Style.DIM)).strip().lower()
        if not ans:
            return "none"
        if ans in {"p", "print"}:
            return "print"
        if ans in {"s", "save"}:
            return "save"
        if ans in {"n", "no", "none", "skip"}:
            return "none"
        print(c("Please answer print, save, or skip.", Style.RED))


def _parse_args(argv: list[str] | None = None) -> argparse.Namespace:
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--verbose",
        action="store_true",
        help="Show a subtle indicator when an opening book move is used.",
    )
    parser.add_argument(
        "--book",
        metavar="PATH",
        default="book.bin",
        help="Path to a Polyglot opening book (default: book.bin).",
    )
    parser.add_argument(
        "--book-chance",
        metavar="P",
        type=float,
        default=0.5,
        help="Chance to use an opening book move when available (0.0-1.0).",
    )
    return parser.parse_args(argv)


def main(argv: list[str] | None = None):
    args = _parse_args(argv)
    print(c("\nBlindfold Chess\n", Style.BOLD))
    print_help()
    print()

    game: Game | None = None

    while True:
        # If a game ended, optionally print PGN, then return to lobby.
        if game is not None and game.ended:
            if game.pgn_text:
                action = ask_pgn_action()
                if action == "print":
                    print()
                    print(c("Final PGN:", Style.CYAN, Style.BOLD))
                    print(finalize_pgn(game.pgn_text, game.player_color, game.engine_name))
                elif action == "save":
                    stamp = datetime.now().strftime("%Y%m%d_%H%M%S")
                    opponent = _slugify_filename(game.engine_name)
                    filename = f"ch0_{opponent}_{stamp}.pgn"
                    with open(filename, "w", encoding="utf-8") as handle:
                        handle.write(
                            str(finalize_pgn(game.pgn_text, game.player_color, game.engine_name))
                        )
                        handle.write("\n")
                    print(c(f"Saved PGN to {filename}", Style.DIM))
            game.close_engine()
            game = None
            print()
            print(c("Lobby. Type 'start' or 'quick' to play.", Style.DIM))
            continue

        # No active game: only limited commands work.
        if game is None:
            user_in = input(c("> ", Style.DIM)).strip()
            if not user_in:
                continue

            cmd = parse_command(user_in)
            if cmd == "start":
                engine_kind, engine_name, engine = choose_engine()
                player_color = choose_color()
                game = Game(
                    engine_kind,
                    engine_name,
                    player_color,
                    engine=engine,
                    verbose=args.verbose,
                    book_path=args.book,
                    book_chance=args.book_chance,
                )

                print()
                print(
                    c("You:", Style.DIM)
                    + " "
                    + c(bool_color_to_string(player_color), Style.CYAN, Style.BOLD)
                    + c(" vs ", Style.DIM)
                    + c(engine_name, Style.CYAN, Style.BOLD)
                )
                print(c(_book_status_line(args.book, args.book_chance), Style.DIM))
                print(c("Tip: type 'show' to display the board.", Style.DIM))
                print()

                # If the engine is white, let it move first.
                if player_color == chess.BLACK:
                    bot_makes_a_move(game)
                continue

            if cmd == "quick":
                engine_kind = "sunfish"
                engine_name = "sunfish"
                engine = None
                player_color = random.choice([chess.WHITE, chess.BLACK])
                game = Game(
                    engine_kind,
                    engine_name,
                    player_color,
                    engine=engine,
                    verbose=args.verbose,
                    book_path=args.book,
                    book_chance=args.book_chance,
                )

                print()
                print(
                    c("You:", Style.DIM)
                    + " "
                    + c(bool_color_to_string(player_color), Style.CYAN, Style.BOLD)
                    + c(" vs ", Style.DIM)
                    + c(engine_name, Style.CYAN, Style.BOLD)
                )
                print(c(_book_status_line(args.book, args.book_chance), Style.DIM))
                print(c("Tip: type 'show' to display the board.", Style.DIM))
                print()

                if player_color == chess.BLACK:
                    bot_makes_a_move(game)
                continue

            if cmd == "help":
                print_help()
                print()
                continue

            if cmd == "quit":
                print(c("Goodbye.", Style.DIM))
                if game is not None:
                    game.close_engine()
                break

            print(c("No active game. Type 'start' or 'quick' (or 'help', 'quit').", Style.RED))
            continue

        # If it's engine's turn, just let it move.
        if game.board.turn != game.player_color:
            bot_makes_a_move(game)
            continue

        turn_prompt = f"{game.board.fullmove_number}> "
        user_in = input(c(turn_prompt, Style.DIM)).strip()
        if not user_in:
            continue

        cmd = parse_command(user_in)

        # Known in-game commands
        if cmd in {"help", "show", "moves", "fen", "pgn", "resign", "quit", "start"}:
            if cmd == "help":
                print_help()
            elif cmd == "show":
                print(game.board)
            elif cmd == "moves":
                moves_in_uci = list(game.board.legal_moves)
                moves_in_san = [game.board.san(m) for m in moves_in_uci]
                print(" ".join(moves_in_san))
            elif cmd == "fen":
                print(game.board.fen())
            elif cmd == "pgn":
                print(finalize_pgn(game.pgn_text, game.player_color, game.engine_name))
            elif cmd == "resign":
                print(c("Resigned.", Style.YELLOW, Style.BOLD))
                result = "0-1" if game.player_color == chess.WHITE else "1-0"
                game.pgn_text += (
                    f" {{ {bool_color_to_string(game.player_color)} resigns. }} {result}"
                )
                game.ended = True
            elif cmd == "quit":
                print(c("Goodbye.", Style.DIM))
                game.close_engine()
                break
            elif cmd == "start":
                print(c("Game in progress. Finish or resign first.", Style.RED))
            continue

        # Otherwise, try to interpret it as a move in SAN
        try:
            game.board.push_san(user_in)
        except ValueError:
            print(c("Illegal move / unknown command.", Style.RED))
            continue

        # Optional: extremely subtle acknowledgement (comment out if you want *zero* noise)
        # print(c("✓", Style.GREEN, Style.DIM))

        if game.turn == chess.WHITE:
            game.count += 1
            game.pgn_text += f"\n{game.count}. {user_in}"
        else:
            game.pgn_text += f" {user_in}"

        check_draw, draw_type = is_a_draw(game.board)
        if check_draw:
            print(c(f"Draw: {draw_type}", Style.YELLOW, Style.BOLD))
            game.pgn_text += " { The game is a draw. } 1/2-1/2"
            game.ended = True
            continue

        if game.board.is_checkmate():
            print(c("Checkmate. You win.", Style.GREEN, Style.BOLD))
            result = "1-0" if game.player_color == chess.WHITE else "0-1"
            game.pgn_text += (
                f" {{ {bool_color_to_string(game.player_color)} wins by checkmate. }} "
                f"{result}"
            )
            game.ended = True
            continue

        game.turn = not game.turn
        # engine will move in the next iteration


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import os
import re
import sys
import time
import subprocess
import signal
import argparse
import importlib
import itertools
import random
import warnings
import chess
import chess.engine
import pathlib
import tqdm
import asyncio
import collections


random.seed(42)


class Command:
    @classmethod
    def add_arguments(cls, parser):
        raise NotImplementedError

    @classmethod
    async def run(cls, engine, args):
        raise NotImplementedError


###############################################################################
# Perft test
###############################################################################

from chess.engine import BaseCommand, UciProtocol


async def uci_perft(engine, depth):
    class UciPerftCommand(BaseCommand[UciProtocol, None]):
        def __init__(self, engine: UciProtocol):
            super().__init__(engine)
            self.moves = []

        def start(self, engine: UciProtocol) -> None:
            engine.send_line(f"go perft {depth}")

        def line_received(self, engine: UciProtocol, line: str) -> None:
            match = re.match("(\w+): (\d+)", line)
            if match:
                move = chess.Move.from_uci(match.group(1))
                cnt = int(match.group(2))
                self.moves.append((move, cnt))

            match = re.match("Nodes searched: (\d+)", line)
            if match:
                self.result.set_result(self.moves)
                self.set_finished()

    return await engine.communicate(UciPerftCommand)


class Perft(Command):
    name = "perft"
    help = "tests for correctness and speed of move generator."

    @classmethod
    def add_arguments(cls, parser):
        parser.add_argument(
            "file", type=argparse.FileType("r"), help="such as tests/queen.fen."
        )
        parser.add_argument("--depth", type=int, default=3)

    @classmethod
    async def run(cls, engine, args):
        lines = args.file.readlines()
        for d in range(1, args.depth + 1):
            if not args.quiet:
                print(f"Going to depth {d}/{args.depth}")

            for line in tqdm.tqdm(lines):
                board, opts = chess.Board.from_epd(line)
                engine._position(board)
                moves = await uci_perft(engine, d)

                cnt = sum(c for m, c in moves)
                opt_cnt = int(opts[f"D{d}"])

                # TODO: Also test that the _number_ of different moves is correct

                if cnt != opt_cnt:
                    print("=========================================")
                    print(f"ERROR at depth {d}. Gave {cnt} rather than {opt_cnt}")
                    print("=========================================")
                    print(board)
                    for m, c in moves:
                        print(m, c)
                    break


class Bench(Command):
    name = "bench"
    help = """Run through a fen file, search every position to a certain depth."""

    @classmethod
    def add_arguments(self, parser):
        parser.add_argument(
            "file", type=argparse.FileType("r"), help="such as tests/mate{1,2,3}.fen."
        )
        parser.add_argument(
            "--depth",
            type=int,
            default=100,
            help="Maximum plies at which to find the mate",
        )
        parser.add_argument(
            "--limit", type=int, default=10000, help="Maximum positions to analyse"
        )

    @classmethod
    async def run(self, engine, args):
        # Run through
        limit = chess.engine.Limit(depth=args.depth)
        lines = args.file.readlines()
        lines = lines[: args.limit]

        total_nodes = 0
        start = time.time()

        pb = tqdm.tqdm(lines)
        for line in pb:
            board, _ = chess.Board.from_epd(line)
            with await engine.analysis(board, limit) as analysis:
                async for info in analysis:
                    pb.set_description(info_to_desc(info))
            total_nodes += info.get("nodes", 0)

        print(f"Total nodes: {total_nodes}.")
        print(f"Average knps: {round(total_nodes/(time.time() - start)/1000, 2)}.")


###############################################################################
# Self-play
###############################################################################


class SelfPlay(Command):
    name = "self-play"
    help = "make sure the engine can complete a game without crashing."

    @classmethod
    def add_arguments(cls, parser):
        parser.add_argument("--time", type=int, default=4000, help="start time in ms")
        parser.add_argument("--inc", type=int, default=100, help="increment in ms")

    @classmethod
    async def run(cls, engine, args):
        wtime = btime = args.time / 1000
        inc = args.inc / 1000
        board = chess.Board()
        with tqdm.tqdm(total=100) as pbar:
            while not board.is_game_over():
                limit = chess.engine.Limit(white_clock=wtime, black_clock=btime, white_inc=inc, black_inc=inc)

                start = time.time()
                result = await engine.play(board, limit)
                elasped = time.time() - start

                if board.turn == chess.WHITE:
                    wtime -= elasped - inc
                else:
                    btime -= elasped - inc

                board.push(result.move)
                pbar.update(1)
            pbar.update(100 - pbar.n)


###############################################################################
# Find mate test
###############################################################################


def info_to_desc(info):
    desc = []
    if "nodes" in info and "time" in info:
        # Add 1 to denominator, since time could be rounded to 0
        nps = info["nodes"] / (info["time"] + 1)
        desc.append(f"knps: {round(nps/1000, 2)}")
    if "depth" in info:
        desc.append(f"depth: {info['depth']}")
    if "score" in info:
        #:wprint(dir(info['score']))
        desc.append(f"score: {info['score'].pov(chess.WHITE).cp/100:.1f}")
    return ", ".join(desc)


def add_limit_argument(parser):
    parser.add_argument(
        "--depth",
        dest="limit_depth",
        type=int,
        default=0,
        help="Maximum plies at which to find the move",
    )
    parser.add_argument(
        "--mate-depth",
        dest="limit_mate",
        type=int,
        default=0,
        help="Maximum plies at which to find the mate",
    )
    parser.add_argument(
        "--movetime",
        dest="limit_movetime",
        type=int,
        default=100,
        help="Movetime in ms",
    )


def get_limit(args):
    if args.limit_depth:
        return chess.engine.Limit(depth=args.limit_depth)
    elif args.limit_mate:
        return chess.engine.Limit(mate=args.limit_mate)
    elif args.limit_movetime:
        return chess.engine.Limit(time=args.limit_movetime / 1000)


class Mate(Command):
    name = "mate"
    help = "Find the mates"

    @classmethod
    def add_arguments(cls, parser):
        parser.add_argument(
            "file", type=argparse.FileType("r"), help="such as tests/mate{1,2,3}.fen."
        )
        parser.add_argument(
            "--limit",
            type=int,
            default=10000,
            help="Take only this many lines from the file",
        )
        add_limit_argument(parser)

    @classmethod
    async def run(cls, engine, args):
        limit = get_limit(args)
        total = 0
        success = 0
        lines = args.file.readlines()
        lines = lines[: args.limit]
        pb = tqdm.tqdm(lines)
        for line in pb:
            board, _ = chess.Board.from_epd(line)
            with await engine.analysis(board, limit) as analysis:
                async for info in analysis:
                    pb.set_description(info_to_desc(info))
                    if not "score" in info:
                        continue
                    score = info["score"]
                    if score.is_mate() or score.relative.cp > 10000:
                        if "pv" in info and info["pv"]:
                            b = board.copy()
                            for move in info["pv"]:
                                b.push(move)
                            if not b.is_game_over():
                                if args.debug:
                                    print("Got mate score, but PV is not mate...")
                                continue
                        if args.debug:
                            print("Found it!")
                        success += 1
                        break
                else:
                    if not args.quiet:
                        print("Failed on", line)
                        print("Result:", info)
        print(f"Succeeded in {success}/{len(lines)} cases.")


class Draw(Command):
    name = "draw"
    help = "Find the draws"

    @classmethod
    def add_arguments(cls, parser):
        parser.add_argument(
            "file", type=argparse.FileType("r"), help="such as tests/stalemate2.fen."
        )
        add_limit_argument(parser)

    @classmethod
    async def run(cls, engine, args):
        limit = get_limit(args)
        total, success = 0, 0
        cnt = collections.Counter()
        pb = tqdm.tqdm(args.file.readlines())
        for line in pb:
            total += 1
            board, _ = chess.Board.from_epd(line)
            with await engine.analysis(board, limit) as analysis:
                last_lower = -10**10
                last_upper = 10**10
                async for info in analysis:
                    pb.set_description(info_to_desc(info))
                    if not "score" in info:
                        continue
                    score = info["score"]
                    if score.is_mate():
                        continue
                    if info.get('lowerbound'):
                        last_lower = score.relative.cp
                    elif info.get('upperbound'):
                        last_upper = score.relative.cp
                    elif score.relative.cp == 0:
                        success += 1
                        cnt[info["depth"]] += 1
                        break
                    if -30 < last_lower and last_upper < 30:
                        success += 1
                        cnt[info["depth"]] += 1
                        break
                else:
                    if not args.quiet:
                        print("Failed on", line.strip())
                        print("Result:", info, 'lower', last_lower, 'upper', last_upper)
                        pass
        print(f"Succeeded in {success}/{total} cases.")
        if not args.quiet:
            print("Depths:")
            for depth, c in cnt.most_common():
                print(f"{depth}: {c}")


###############################################################################
# Best move test
###############################################################################


class Best(Command):
    name = "best"
    help = "Find the best move"

    @classmethod
    def add_arguments(cls, parser):
        parser.add_argument(
            "file", type=argparse.FileType("r"), help="such as bratko_kopec_test.epd."
        )
        parser.add_argument(
            "--limit",
            type=int,
            default=10000,
            help="Take only this many lines from the file",
        )
        add_limit_argument(parser)

    @classmethod
    async def run(cls, engine, args):
        limit = get_limit(args)
        points, total = 0, 0
        lines = args.file.readlines()
        # random.shuffle(lines)
        lines = lines[: args.limit]
        for line in (pb := tqdm.tqdm(lines)):
            board, opts = chess.Board.from_epd(line)
            if "pv" in opts:
                for move in opts["pv"]:
                    board.push(move)
            # When using a pv I need to supply bm/am using a comment, because
            # otherwise python-chess won't parse it
            if "c0" in opts:
                # The comment format is expected to be like 'am f1f2; bm f2f3'
                for key, val in re.findall("(\w+) (\w+)", opts["c0"]):
                    opts[key] = [chess.Move.from_uci(val)]
            if "am" not in opts and "bm" not in opts:
                if not args.quiet:
                    print("Line didn't have am/bm in opts", line, opts)
                continue
            # am -> avoid move; bm -> best move
            pb.set_description(opts.get("id", ""))
            result = await engine.play(board, limit, info=chess.engine.INFO_SCORE)
            errors = []
            if "bm" in opts:
                total += 1
                if result.move in opts["bm"]:
                    points += 1
                else:
                    errors.append(f'Gave move {result.move} rather than {opts["bm"]}')
            if "am" in opts:
                total += 1
                if result.move not in opts["am"]:
                    points += 1
                else:
                    errors.append(f'Gave move {result.move} which is in {opts["am"]}')
            if not args.quiet and errors:
                print("Failed on", line.strip())
                for er in errors:
                    print(er)
                print("Full result:", result)
                print()
            pb.set_postfix(acc=points / total)
        print(f"Succeeded in {points}/{total} cases.")


###############################################################################
# Actions
###############################################################################


def main():
    parser = argparse.ArgumentParser(
        description="Run various tests for speed and correctness of sunfish."
    )
    parser.add_argument("args", help="Command and arguments to run")
    parser.add_argument(
        "--debug", action="store_true", help="Write lots of extra stuff"
    )
    parser.add_argument("--quiet", action="store_true", help="Only write pass/fail")
    parser.add_argument(
        "--xboard", action="store_true", help="Use xboard protocol instead of uci"
    )
    subparsers = parser.add_subparsers()
    subparsers.required = True

    for cls in Command.__subclasses__():
        sub = subparsers.add_parser(cls.name, help=cls.help)
        cls.add_arguments(sub)
        sub.set_defaults(func=cls.run)

    args = parser.parse_args()

    if args.debug:
        import logging

        logging.basicConfig(level=logging.DEBUG)

    async def run():
        if args.xboard:
            _, engine = await chess.engine.popen_xboard(args.args.split())
        else:
            _transport, engine = await chess.engine.popen_uci(args.args.split())
        try:
            await args.func(engine, args)
        finally:
            await engine.quit()

    asyncio.set_event_loop_policy(chess.engine.EventLoopPolicy())
    start = time.time()
    asyncio.run(run())
    print(f"Took {round(time.time() - start, 2)} seconds.")


if __name__ == "__main__":
    main()
